package io.irandom.profiling;

import java.util.Vector;
import salsa.language.Actor;
import io.core.profiling.DecisionAgent;
import io.core.language.*;
import io.core.messaging.*;
import io.core.profiling.*;
import io.random.messaging.RandomDecisionAgent;
import io.irandom.protocol.IRandomSteal;
import io.irandom.messaging.IRandomPeerTheater;

public class IRandomDecisionAgent extends RandomDecisionAgent implements DecisionAgent {

	//Waiting determines if the decision agent will originate another random stealing packet.
	private boolean waiting = false;
	private IOTheater theater;

	private long lastReset;

	public IRandomDecisionAgent(IOTheater _theater) {
		theater = _theater;
		lastReset = System.currentTimeMillis();
	}

	public void stealTerminated() {
		waiting = false;
	}

	public void reset() {
		//reset all the autonomous actors in the system
		lastReset = System.currentTimeMillis();

		Vector registry = new Vector(theater.getRegistry().values());

		for (int i = 0; i < registry.size(); i++) {
			if (registry.get(i) instanceof IOActor) {
				((IOActor)registry.get(i)).reset();
			}
		}
	}

	private double getProcessing() {
		return ((IRandomPeerTheater)theater).getProcessing();
	}

	private double getBandwidth() {
		return ((IRandomPeerTheater)theater).getBandwidth();
	}

	private double decisionFunction(OfferedData foreignData, ActorData actorData, TheaterLocator target) {
		double foreignProcessed = 0;

		if (actorData.processed > 0) {
			//Find the average cycles per message processed
			double cycles = actorData.processingTime * getProcessing() / actorData.processed;

			//Find the time it will take to process single messages on the foreign theater
			double foreignProcessTime = cycles / foreignData.processing;

			double foreignTotalProcessTime = 0;

			if (actorData.sentForeign > 0) {
				//Find the average bits per message sent to a foreign theater
				double bits = (actorData.sendingTime * getBandwidth()) / actorData.sentForeign;

				//Find the time it will take to send single messages on the foreign theater
				double foreignSendTime = bits / foreignData.bandwidth;

				//To fully process one message (including sending what it sends) on foreign theater
				double sendingRatio = actorData.messagesNotSentTo(target) / actorData.processed;

				foreignTotalProcessTime = (foreignProcessTime + (sendingRatio * foreignSendTime));
			} else {
				foreignTotalProcessTime = foreignProcessTime;
			}

			foreignProcessed = actorData.collectionTime / foreignTotalProcessTime;
		} else {
			foreignProcessed = actorData.recieved;
		}

		return foreignProcessed - actorData.processed;
	}

	public IOActor getBestCandidate(OfferedData offeredData, TheaterLocator target) {
		//returns the best candidate actor to migrate to the given theater

		Vector registry = new Vector(theater.getRegistry().values());

		IOActor candidate = null;
		double candidateGain = -1;

		for (int i = 0; i < registry.size(); i++) {
			if (registry.get(i) instanceof IOActor) {
				AutonomousActor current = (AutonomousActor)registry.get(i);

				if (( current.local ) && ( !current.scheduled() )) {
					double gain = decisionFunction(offeredData, current.getData(), target);

					if ((!current.satisfied()) && (gain > candidateGain)) {
						candidate = (IOActor)current;
						candidateGain = gain;
					}
				}
			}
		}

		return candidate;
	}

	private OfferedData getTheaterData() {
		Vector registry = new Vector(theater.getRegistry().values());

		long sendingUsed = 0;
		long processingUsed = 0;

		for (int i = 0; i < registry.size(); i++) {
			if ((registry.get(i) instanceof IOActor) && ( ((Actor)registry.get(i)).actorState != null )) {
				sendingUsed += ((IOActor)registry.get(i)).getData().sendingTime;
				processingUsed += ((IOActor)registry.get(i)).getData().processingTime;
			}
		}

		long collectionTime = System.currentTimeMillis() - lastReset;

		return new OfferedData(  (1 - ((getBandwidth() * sendingUsed)/collectionTime)),
					 (1 - ((getProcessing() * processingUsed)/collectionTime))
				      );
	}

	private void beginSteal() {
		if (!theater.getNeighbors().isEmpty()) {
			System.err.println("Sending random steal packet");
			waiting = true;

			//create a RandomSteal packet, and send it to a random neighbor
			IRandomSteal iRandomSteal = new IRandomSteal(theater.getLocation(), getTheaterData());
			iRandomSteal.randomSend(theater.getNeighbors());
		}
	}

	private boolean overloaded() {
		Vector registry = new Vector(theater.getRegistry().values());

		for (int i = 0; i < registry.size(); i++) {
			if (registry.get(i) instanceof IOActor) {
				if ( !((IOActor)registry.get(i)).satisfied() ) {
					return true;
				}
			}
		}

		return false;
	}

	private void printStatistics() {
		Vector registry = new Vector(theater.getRegistry().values());

		int messagesSent = 0;
		int messagesRecieved = 0;
		int messagesProcessed = 0;

		for (int i = 0; i < registry.size(); i++) {
			if (registry.get(i) instanceof IOActor) {
				IOActor current = (IOActor)registry.get(i);

				messagesSent += current.getData().sentLocal;
				messagesSent += current.getData().sentForeign;

				messagesRecieved += current.getData().recieved;
				messagesProcessed += current.getData().processed;
			}
		}

		System.out.println(System.currentTimeMillis() + "\t" + messagesSent + "\t" + 
				   messagesProcessed + "\t\t" + messagesRecieved);
	}

	public void run() {
		System.out.println("Time\t\tSent\tProcessed\tRecieved");

		//loop forever
		while (true) {
			try {
				sleep(10000);
			} catch (InterruptedException ie) {
				System.err.println("Sleep interrupted!\n");
			}

			if (!overloaded() && !waiting) {
				beginSteal();
			}

			printStatistics();
			reset();
		}
	}
}
